import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function
from torch.nn.modules.utils import _pair
from torch.nn import init
import unoptimized_cuda_kernel
import math
import numpy as np
import time


class UnoptimizedLinear(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
 
        super(UnoptimizedLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            # You should always register all possible parameters, but the
            # optional ones can be None if you want.
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            init.uniform_(self.bias, -bound, bound)

    def forward(self, input):   
        # start_time = time.time()
        out = torch.zeros([input.size(0),self.weight.size(0)], dtype=torch.float, device=torch.device('cuda:0'))
        if self.bias is not None:
            unoptimized_cuda_kernel.UNOPTIMIZED_LINEAR(input, self.weight, self.bias,out)
        else:
            temp = torch.zeros([self.weight.size(0)], dtype=torch.float, device=torch.device('cuda:0'))
            unoptimized_cuda_kernel.UNOPTIMIZED_LINEAR(input, self.weight, temp,out)
        # end_time = time.time()
        # print("Linear Time:", end_time - start_time )   
        return out

       

    def extra_repr(self):
        # (Optional)Set the extra information about this module. You can test
        # it by printing an object of this class.
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None  
        )
